@@ -1,8 +1,11 @@
+import sys
+
 # stdlib
 from operator import itemgetter
 from functools import lru_cache
 
 # 3rd party
+import arrow
 import meilisearch
 from django.apps import apps
 from django.db.models import Q, Case, When, Model, Manager, QuerySet
@@ -10,12 +13,13 @@ from wagtail.search.index import (
     FilterField, SearchField, RelatedFields, AutocompleteField, class_is_indexed,
     get_indexed_models
 )
-from wagtail.search.utils import OR, AND
+from wagtail.search.utils import OR
 from django.utils.encoding import force_text
 from wagtail.search.backends.base import (
     BaseSearchBackend, BaseSearchResults, EmptySearchResults, BaseSearchQueryCompiler
 )
 
+
 from .settings import STOP_WORDS
 
 try:
@@ -60,6 +64,11 @@ class MeiliSearchModelIndex:
             'limit': 999999,
             'matches': 'true'
         }
+        self.update_strategy = backend.update_strategy
+        self.update_delta = backend.update_delta
+        self.delta_fields = [
+            'created_at', 'updated_at', 'first_published_at', 'last_published_at'
+        ]
 
     def _set_index(self, model):
         label = self._get_label(model)
@@ -129,19 +138,30 @@ class MeiliSearchModelIndex:
         """
         for field in model.get_search_fields():
             if isinstance(field, (SearchField, FilterField, AutocompleteField)):
-                yield _get_field_mapping(field), self.prepare_value(field.get_value(item))
+                try:
+                    yield _get_field_mapping(field), self.prepare_value(field.get_value(item))
+                except Exception:
+                    pass
             if isinstance(field, RelatedFields):
                 value = field.get_value(item)
                 if isinstance(value, (Manager, QuerySet)):
                     qs = value.all()
                     for sub_field in field.fields:
                         sub_values = qs.values_list(sub_field.field_name, flat=True)
-                        yield '{0}__{1}'.format(field.field_name, _get_field_mapping(sub_field)), \
-                            self.prepare_value(list(sub_values))
+                        try:
+                            yield '{0}__{1}'.format(
+                                field.field_name, _get_field_mapping(sub_field)), \
+                                self.prepare_value(list(sub_values))
+                        except Exception:
+                            pass
                 if isinstance(value, Model):
                     for sub_field in field.fields:
-                        yield '{0}__{1}'.format(field.field_name, _get_field_mapping(sub_field)),\
-                            self.prepare_value(sub_field.get_value(value))
+                        try:
+                            yield '{0}__{1}'.format(
+                                field.field_name, _get_field_mapping(sub_field)),\
+                                self.prepare_value(sub_field.get_value(value))
+                        except Exception:
+                            pass
 
     @lru_cache()
     def _create_document(self, model, item):
@@ -167,8 +187,17 @@ class MeiliSearchModelIndex:
         pass
 
     def add_item(self, item):
+        if self.update_strategy == 'delta':
+            # We send it a list and get back a list, though that list might be empty
+            checked = self._check_deltas([item, ])
+            if len(checked):
+                item = checked[0]
+
         doc = self._create_document(self.model, item)
-        self.index.add_documents([doc])
+        if self.update_strategy == 'soft':
+            self.index.update_documents([doc])
+        else:
+            self.index.add_documents([doc])
 
     def add_items(self, item_model, items):
         """Adds items in bulk to the index. If we're adding stuff through the `update_index`
@@ -196,17 +225,54 @@ class MeiliSearchModelIndex:
 
         # split items into chunks of 100
         chunks = [items[x:x + 100] for x in range(0, len(items), 100)]
+
         for chunk in chunks:
+            if self.update_strategy == 'delta':
+                chunk = self._check_deltas(chunk)
             prepared = []
             for item in chunk:
                 doc = self._create_document(self.model, item)
                 prepared.append(doc)
 
-            self.index.add_documents(prepared)
+            if self.update_strategy == 'soft' or self.update_strategy == 'delta':
+                self.index.update_documents(prepared)
+            else:
+                self.index.add_documents(prepared)
             del(chunk)
 
         return True
 
+    def _has_date_fields(self, obj):
+        find = self.delta_fields
+        fields = [_.name for _ in obj._meta.fields]
+        rv = any(item in find for item in fields)
+        return rv
+
+    def _check_deltas(self, objects: list) -> list:
+        """Takes a list of objects and removes any where the last_published_at, first_published_at,
+        created_at or updated_at are outside of the time delta.
+
+        TODO: This looks ugly, and is probably slow.
+
+        Args:
+            objects (list): A list of model instances
+        """
+        filtered = []
+        since = arrow.now().shift(**self.update_delta).datetime
+        for obj in objects:
+            if self._has_date_fields(obj):
+                for field in self.delta_fields:
+                    if hasattr(obj, field):
+                        val = getattr(obj, field)
+                        try:
+                            if val and val > since:
+                                filtered.append(obj)
+                                continue
+                        except TypeError:
+                            pass
+
+        return filtered
+
     def delete_item(self, obj):
         self.index.delete_document(obj.id)
 
@@ -217,15 +283,53 @@ class MeiliSearchModelIndex:
         return self.name
 
 
+class DummyModelIndex:
+
+    """This class enables the SKIP_MODELS feature by providing a
+    dummy model index that we can add things to without it actually
+    doing anything.
+    """
+
+    def add_model(self, model):
+        pass
+
+    def add_items(self, model, chunk):
+        pass
+
+
 class MeiliSearchRebuilder:
     def __init__(self, model_index):
         self.index = model_index
         self.uid = self.index._get_label(self.index.model)
+        self.dummy_index = DummyModelIndex()
 
     def start(self):
-        # for now, we're going to just delete every document in the passed index
-        old_index = self.index.backend.client.get_index(self.uid)
-        old_index.delete_all_documents()
+        """This is the thing that starts of a rebuild of the search
+        index. We offer three strategies, `hard`, `soft` and `delta`.
+
+        * `hard` will delete every document in the index and try to add them anew
+        * `soft` will do an "add or update" for each document
+        * `delta` will attempt to only update documents that have been saved in the
+            last X amount of time
+
+        The trade off here is that a `hard` update is CPU intensive for quite a long time, while
+        a `soft` update can leave fields in existing indexed documents that aren't in the new
+        document. Once a large site is fully indexed, it should be pretty safe to switch to a
+        `delta` strategy which would be the least CPU intensive of all.
+        """
+        if self.index.model._meta.label in self.index.backend.skip_models:
+            sys.stdout.write(f'SKIPPING: {self.index.model._meta.label}\n')
+            return self.dummy_index
+
+        strategy = self.index.backend.update_strategy
+        if strategy == 'soft' or strategy == 'delta':
+            # SOFT UPDATE STRATEGY
+            index = self.index.backend.client.get_index(self.uid)
+        else:
+            # HARD UPDATE STRATEGY
+            old_index = self.index.backend.client.get_index(self.uid)
+            old_index.delete_all_documents()
+
         model = self.index.model
         index = self.index.backend.get_index_for_model(model)
         return index
@@ -278,6 +382,15 @@ def get_descendant_models(model):
 class MeiliSearchResults(BaseSearchResults):
     supports_facet = False
 
+    def _get_field_boosts(self, model):
+        boosts = {}
+        for field in model.search_fields:
+            if isinstance(field, SearchField):
+                if hasattr(field, 'boost'):
+                    boosts[field.field_name] = field.boost
+
+        return boosts
+
     def _do_search(self):
         results = []
 
@@ -289,8 +402,10 @@ class MeiliSearchResults(BaseSearchResults):
         for m in models:
             index = self.backend.get_index_for_model(m)
             rv = index.search(terms)
+            boosts = self._get_field_boosts(m)
             for item in rv['hits']:
                 if item not in results:
+                    item['boosts'] = boosts
                     results.append(item)
 
         """At this point we have a list of results that each look something like this
@@ -298,6 +413,9 @@ class MeiliSearchResults(BaseSearchResults):
 
         {
             'id': 45014,
+            'boosts': {
+                'title': 10
+            },
             '_matchesInfo': {
                 'title_filter': [
                     {'start': 0, 'length': 6}
@@ -327,10 +445,20 @@ class MeiliSearchResults(BaseSearchResults):
         # The simplest way is probably to len(str(item['_matchesInfo'])) which for the
         # above example returns a score of 386 and for the bottom result in my test set is
         # just 40.
-
-        # TODO: Implement `boost` on fields
         for item in results:
-            item['score'] = len(str(item['_matchesInfo']))
+            score = 0
+            for key in item['_matchesInfo']:
+                try:
+                    boost = item['boosts'].get(key, 1)
+                except Exception:
+                    boost = 1
+
+                if not boost:
+                    boost = 1
+
+                score += len(str(item['_matchesInfo'][key])) * boost
+
+            item['score'] = score
 
         sorted_results = sorted(results, key=itemgetter('score'), reverse=True)
         sorted_ids = [_['id'] for _ in sorted_results]
@@ -369,6 +497,11 @@ class MeiliSearchBackend(BaseSearchBackend):
         except Exception:
             raise
         self.stop_words = params.get('STOP_WORDS', STOP_WORDS)
+        self.skip_models = params.get('SKIP_MODELS', [])
+        self.update_strategy = params.get('UPDATE_STRATEGY', 'soft')
+        self.update_delta = None
+        if self.update_strategy == 'delta':
+            self.update_delta = params.get('UPDATE_DELTA', {'weeks': -1})
 
     def _refresh(self, uid, model):
         index = self.client.get_index(uid)
